<div class="card d-none d-md-block mt-4">
    <h5 class="card-header font-weight-bold">Version {{ release }}</h5>
    <div class="card-body">
        <ul class="list-group list-group-flush list-unstyled quicklinks">
            <li><a href="{{ pathto('auto_tutorials/index.html', 1) }}"><i class="text-muted fas fa-book fa-fw"></i> Tutorials</a></li>
            <li><a href="{{ pathto('whats_new.html', 1) }}"><i class="text-muted fas fa-newspaper fa-fw"></i> Changelog</a></li>
            <li><a href="{{ pathto('overview/get_help.html', 1) }}"><i class="text-muted fas fa-circle-question fa-fw"></i> Get help</a></li>
            <li><a href="{{ pathto('overview/cite.html', 1) }}"><i class="text-muted fas fa-quote-left fa-fw"></i> Cite</a></li>
            <li><a href="{{ pathto('install/contributing.html', 1) }}"><i class="text-muted fas fa-code-branch fa-fw"></i> Contribute</a></li>
        </ul>
    </div>
</div>
