</div>
</div>
<footer>
  <nav class="navbar fixed-bottom navbar-light bg-light border-top justify-content-center pt-0 pb-0 small">
    <span>Created on {{ date }} via <a href="https://mne.tools" target="_blank">MNE-Python</a> {{ mne_version }}</span>
  </nav>
</footer>
</body>

</html>
