.. _changes_0_3:

Version 0.3 (2012-03-23)
------------------------

Changelog
~~~~~~~~~

- Sign flip computation for robust label average of signed values by `Alex Gramfort`_.

- Reading and writing of .w files by `Martin Luessi`_.

- Support for modifying Raw object and allow raw data preloading with memory mapping by `Martin Luessi`_ and `Alex Gramfort`_.

- Support of arithmetic of Evoked data (useful to concatenate between runs and compute contrasts) by `Alex Gramfort`_.

- Support for computing sensor space data from a source estimate using an MNE forward solution by `Martin Luessi`_.

- Support of arithmetic of Covariance by `Alex Gramfort`_.

- Write BEM surfaces in Python  by `Alex Gramfort`_.

- Filtering operations and apply_function interface for Raw object by `Martin Luessi`_.

- Support for complex valued raw fiff files and computation of analytic signal for Raw object by `Martin Luessi`_.

- Write inverse operators (surface and volume) by `Alex Gramfort`_.

- Covariance matrix computation with multiple event types by `Martin Luessi`_.

- New tutorial in the documentation and new classes and functions reference page by `Alex Gramfort`_.

Authors
~~~~~~~

The committer list for this release is the following (preceded by number
of commits):

* 80  Alexandre Gramfort
* 51  Martin Luessi
