.. _Abram Hindle: http://softwareprocess.es

.. _Adam Li: http://github.com/adam2392

.. _Adeline Fecker: https://github.com/adelinefecker

.. _Adina Wagner: https://github.com/adswa

.. _Adonay Nunes: https://github.com/AdoNunes

.. _Alan Leggitt: https://github.com/leggitta

.. _Alejandro Weinstein: http://ocam.cl

.. _Alessandro Tonin: https://www.linkedin.com/in/alessandro-tonin-7892b046

.. _Alex Ciok: https://github.com/alexCiok

.. _Alex Gramfort: http://alexandre.gramfort.net

.. _Alex Rockhill: https://github.com/alexrockhill/

.. _Alexander Rudiuk: https://github.com/ARudiuk

.. _Alexandre Barachant: http://alexandre.barachant.org

.. _Andrea Brovelli: https://andrea-brovelli.net

.. _Andreas Hojlund: https://github.com/ahoejlund

.. _Andres Rodriguez: https://github.com/infinitejest/

.. _Andrew Dykstra: https://github.com/adykstra

.. _Andrew Quinn: https://github.com/ajquinn

.. _Aniket Pradhan: https://github.com/Aniket-Pradhan

.. _Anna Padee: https://github.com/apadee/

.. _Annalisa Pascarella: https://www.dima.unige.it/~pascarel/html/cv.html

.. _Anne-Sophie Dubarry: https://github.com/annesodub

.. _Antoine Gauthier: https://github.com/Okamille

.. _Antti Rantala: https://github.com/Odingod

.. _Apoorva Karekal: https://github.com/apoorva6262

.. _Archit Singhal: https://github.com/architsinghal-mriirs

.. _Ashley Drew: https://github.com/ashdrew

.. _Asish Panda: https://github.com/kaichogami

.. _Austin Hurst: https://github.com/a-hurst

.. _Ben Beasley: https://github.com/musicinmybrain

.. _Britta Westner: https://britta-wstnr.github.io

.. _Bruno Nicenboim: https://bnicenboim.github.io

.. _buildqa: https://github.com/buildqa

.. _Carlos de la Torre-Ortiz: https://ctorre.me

.. _Carina Forster: https://github.com/carinafo

.. _Cathy Nangini: https://github.com/KatiRG

.. _Chetan Gohil: https://github.com/cgohil8

.. _Chris Bailey: https://github.com/cjayb

.. _Chris Holdgraf: https://chrisholdgraf.com

.. _Chris Mullins: http://crmullins.com

.. _Christian Brodbeck: https://github.com/christianbrodbeck

.. _Christian O'Reilly: https://github.com/christian-oreilly

.. _Christopher Dinh: https://github.com/chdinh

.. _Chun-Hui Li: https://github.com/iamsc

.. _Clemens Brunner: https://github.com/cbrnr

.. _Cora Kim: https://github.com/kimcoco

.. _Cristóbal Moënne-Loccoz: https://github.com/cmmoenne

.. _Dan Wakeman: https://github.com/dgwakeman

.. _Daniel Carlström Schad: https://github.com/Dod12

.. _Daniel Hasegan: https://daniel.hasegan.com

.. _Daniel McCloy: http://dan.mccloy.info

.. _Daniel Strohmeier: https://github.com/joewalter

.. _Darin Erat Sleiter: https://github.com/dsleiter

.. _David Haslacher: https://github.com/davidhaslacher

.. _David Julien: https://github.com/Swy7ch

.. _David Sabbagh: https://github.com/DavidSabbagh

.. _Demetres Kostas: https://github.com/kostasde

.. _Denis Engemann: http://denis-engemann.de

.. _Dirk Gütlin: https://github.com/DiGyt

.. _Dmitrii Altukhov: https://github.com/dmalt

.. _Dominik Welke: https://github.com/dominikwelke/

.. _Dominik Wetzel: https://github.com/schmetzler

.. _Dominique Makowski: https://dominiquemakowski.github.io/

.. _Eberhard Eich: https://github.com/ebeich

.. _Eduard Ort: https://github.com/eort

.. _Emily Stephen: http://github.com/emilyps14

.. _Enrico Varano: https://github.com/enricovara/

.. _Enzo Altamiranda: https://www.linkedin.com/in/enzoalt

.. _Eric Larson: http://larsoner.com

.. _Erica Peterson: https://github.com/nordme

.. _Erkka Heinila: https://github.com/Teekuningas

.. _Etienne de Montalivet: https://github.com/etiennedemontalivet

.. _Evan Hathaway: https://github.com/ephathaway

.. _Evgeny Goldstein: https://github.com/evgenygoldstein

.. _Ezequiel Mikulan: https://github.com/ezemikulan

.. _Ezequiel Mikulan: https://github.com/ezemikulan

.. _Fahimeh Mamashli: https://github.com/fmamashli

.. _Federico Raimondo: https://github.com/fraimondo

.. _Federico Zamberlan: https://github.com/fzamberlan

.. _Felix Klotzsche: https://github.com/eioe

.. _Felix Raimundo: https://github.com/gamazeps

.. _Frederik Weber: https://github.com/Frederik-D-Weber

.. _Fu-Te Wong: https://github.com/zuxfoucault

.. _Geoff Brookshire: https://github.com/gbrookshire

.. _Guillaume Dumas: http://www.extrospection.eu

.. _Guillaume Favelier: https://github.com/GuillaumeFavelier

.. _Hakimeh Aslsardroud: https://www.researchgate.net/profile/Hakimeh-Pourakbari

.. _Hamid Maymandi: https://github.com/HamidMandi

.. _Hari Bharadwaj: http://www.haribharadwaj.com

.. _Henrich Kolkhorst: https://github.com/hekolk

.. _Hongjiang Ye: https://github.com/rubyyhj

.. _Hubert Banville: https://github.com/hubertjb

.. _Hüseyin Orkun Elmas: https://github.com/HuseyinOrkun

.. _Ilias Machairas: https://github.com/JungleHippo

.. _Ivana Kojcic: https://github.com/ikojcic

.. _Jaakko Leppakangas: https://github.com/jaeilepp

.. _Jack Zhang: https://github.com/jackz314

.. _Jair Montoya Martinez: https://github.com/jmontoyam

.. _Jan Sedivy: https://github.com/honzaseda

.. _Jan Sosulski: https://jan-sosulski.de

.. _Jan Zerfowski: https://github.com/jzerfowski

.. _Jasper van den Bosch: https://github.com/ilogue

.. _Jean-Baptiste Schiratti: https://github.com/jbschiratti

.. _Jean-Remi King: https://github.com/kingjr

.. _Jeff Stout: https://megcore.nih.gov/index.php/Staff

.. _Jeroen Van Der Donckt: https://github.com/jvdd

.. _Jesper Duemose Nielsen: https://github.com/jdue

.. _Jevri Hanna: https://github.com/jshanna100

.. _jeythekey: https://github.com/jeythekey

.. _Joan Massich: https://github.com/massich

.. _Johann Benerradi: https://github.com/HanBnrd

.. _Johannes Niediek: https://github.com/jniediek

.. _John Samuelsson: https://github.com/johnsam7

.. _John Veillette: https://psychology.uchicago.edu/directory/john-veillette

.. _Jon Houck: https://www.mrn.org/people/jon-m.-houck/principal-investigators

.. _Jona Sassenhagen: https://github.com/jona-sassenhagen

.. _Jonathan Kuziek:  https://github.com/kuziekj

.. _Jordan Drew: https://github.com/jadrew43

.. _Jose Alanis: https://github.com/JoseAlanis

.. _Joshua Bear: https://github.com/joshbear

.. _Joshua Teves: https://github.com/jbteves

.. _Judy D Zhu: https://github.com/JD-Zhu

.. _Juergen Dammers: https://github.com/jdammers

.. _Jukka Nenonen: https://www.linkedin.com/pub/jukka-nenonen/28/b5a/684

.. _Jussi Nurminen: https://github.com/jjnurminen

.. _Kaisu Lankinen: http://bishoplab.berkeley.edu/Kaisu.html

.. _kalenkovich: https://github.com/kalenkovich

.. _Katarina Slama: https://github.com/katarinaslama

.. _Keith Doelling: https://github.com/kdoelling1919

.. _Kostiantyn Maksymenko: https://github.com/makkostya

.. _Kyle Mathewson: https://github.com/kylemath

.. _Larry Eisenman:  https://github.com/lneisenman

.. _Lau Møller Andersen: https://github.com/ualsbombe

.. _Laura Gwilliams: http://lauragwilliams.github.io

.. _Leonardo Barbosa: https://github.com/noreun

.. _Liberty Hamilton: https://github.com/libertyh

.. _Lorenzo Desantis: https://github.com/lorenzo-desantis/

.. _Lukas Breuer: https://www.researchgate.net/profile/Lukas-Breuer-2

.. _Lukas Gemein: https://github.com/gemeinl

.. _Lukáš Hejtmánek: https://github.com/hejtmy

.. _Luke Bloy: https://www.research.chop.edu/imaging/team

.. _Lx37: https://github.com/Lx37

.. _Mads Jensen: https://github.com/MadsJensen

.. _Maggie Clarke: https://github.com/mdclarke

.. _Mainak Jas: https://jasmainak.github.io

.. _Marcin Koculak: https://github.com/mkoculak

.. _Marian Dovgialo: https://github.com/mdovgialo

.. _Marijn van Vliet: https://github.com/wmvanvliet

.. _Mark Alexander Henney: https://github.com/henneysq

.. _Mark Wronkiewicz: https://ml.jpl.nasa.gov/people/wronkiewicz/wronkiewicz.html

.. _Marmaduke Woodman: https://github.com/maedoc

.. _Martin Billinger: https://github.com/mbillingr

.. _Martin Luessi: https://github.com/mluessi

.. _Martin Schulz: https://github.com/marsipu

.. _Mathieu Scheltienne: https://github.com/mscheltienne

.. _Mathurin Massias: https://mathurinm.github.io/

.. _Mats van Es: https://github.com/matsvanes

.. _Matt Boggess: https://github.com/mattboggess

.. _Matt Courtemanche: https://github.com/mjcourte

.. _Matt Sanderson: https://github.com/monkeyman192

.. _Matteo Anelli: https://github.com/matteoanelli

.. _Matthias Dold: https://matthiasdold.de

.. _Matthias Eberlein: https://github.com/MatthiasEb

.. _Matti Toivonen: https://github.com/mattitoi

.. _Mauricio Cespedes Tenorio: https://github.com/mcespedes99

.. _Michiru Kaneda: https://github.com/rcmdnk

.. _Mikołaj Magnuski: https://github.com/mmagnuski

.. _Milan Rybář: http://milanrybar.cz

.. _Mingjian He: https://github.com/mh105

.. _Mohammad Daneshzand: https://github.com/mdaneshzand

.. _Moritz Gerster: https://github.com/moritz-gerster

.. _Natalie Klein: https://github.com/natalieklein

.. _Nathalie Gayraud: https://github.com/ngayraud

.. _Naveen Srinivasan: https://github.com/naveensrinivasan

.. _Nick Foti: http://nfoti.github.io

.. _Nick Ward: http://www.ucl.ac.uk/ion/departments/sobell/Research/NWard

.. _Nicolas Barascud: https://github.com/nbara

.. _Niklas Wilming: https://github.com/nwilming

.. _Nikolai Chapochnikov: https://github.com/chapochn

.. _Nikolas Chalas: https://github.com/Nichalas

.. _Okba Bekhelifi: https://github.com/okbalefthanded

.. _Olaf Hauk: http://www.neuroscience.cam.ac.uk/directory/profile.php?olafhauk

.. _Oleh Kozynets: https://github.com/OlehKSS

.. _Pablo-Arias: https://github.com/Pablo-Arias

.. _Padma Sundaram: https://www.nmr.mgh.harvard.edu/user/8071

.. _Paul Pasler: https://github.com/ppasler

.. _Paul Roujansky: https://github.com/paulroujansky

.. _Peter Molfese: https://github.com/pmolfese

.. _Phillip Alday: https://palday.bitbucket.io

.. _Pierre Ablin: https://pierreablin.com

.. _Pierre-Antoine Bannier: https://github.com/PABannier

.. _Qianliang Li: https://www.dtu.dk/english/service/phonebook/person?id=126774

.. _Quentin Bertrand: https://github.com/QB3

.. _Qunxi Dong: https://github.com/dongqunxi

.. _Rahul Nadkarni: https://github.com/rahuln

.. _Ram Pari: https://github.com/ramkpari

.. _Ramiro Gatti: https://github.com/ragatti

.. _ramonapariciog: https://github.com/ramonapariciog

.. _Rasmus Zetter: https://people.aalto.fi/rasmus.zetter

.. _Reza Nasri: https://github.com/0reza

.. _Reza Shoorangiz: https://github.com/rezashr

.. _Richard Höchenberger: https://github.com/hoechenberger

.. _Richard Koehler: https://github.com/richardkoehler

.. _Riessarius Stargardsky: https://github.com/Riessarius

.. _Roan LaPlante: https://github.com/aestrivex

.. _Robert Luke: https://github.com/rob-luke

.. _Robert Seymour: https://neurofractal.github.io

.. _Romain Derollepot: https://github.com/rderollepot

.. _Romain Trachel: https://fr.linkedin.com/in/trachelr

.. _Roman Goj: http://romanmne.blogspot.co.uk

.. _Ross Maddox: https://www.urmc.rochester.edu/labs/maddox-lab.aspx

.. _Rotem Falach: https://github.com/Falach

.. _Samu Taulu: https://phys.washington.edu/people/samu-taulu

.. _Samuel Deslauriers-Gauthier: https://github.com/sdeslauriers

.. _Samuel Powell: https://github.com/samuelpowell

.. _Santeri Ruuskanen: https://github.com/ruuskas

.. _Sara Sommariva: http://www.dima.unige.it/~sommariva/

.. _Scott Huberty:  https://orcid.org/0000-0003-2637-031X

.. _Sebastiaan Mathot: http://www.cogsci.nl/smathot

.. _Sebastian Castano: https://github.com/jscastanoc

.. _Sebastian Major: https://github.com/major-s

.. _Sébastien Marti: https://www.researchgate.net/profile/Sebastien-Marti

.. _Sena Er: https://github.com/sena-neuro

.. _Senwen Deng: https://snwn.de

.. _Sheraz Khan: https://github.com/SherazKhan

.. _Silvia Cotroneo: https://github.com/sfc-neuro

.. _Simeon Wong: https://github.com/dtxe

.. _Simon Kern: https://github.com/skjerns

.. _Simon Kornblith: http://simonster.com

.. _Sondre Foslien: https://github.com/sondrfos

.. _Sophie Herbst: https://github.com/SophieHerbst

.. _Stanislas Chambon: https://github.com/Slasnista

.. _Stefan Appelhoff: http://stefanappelhoff.com

.. _Stefan Repplinger: https://github.com/stfnrpplngr

.. _Steven Bethard: https://github.com/bethard

.. _Steven Bierer: https://github.com/neurolaunch

.. _Steven Gutstein: https://github.com/smgutstein

.. _Sumalyo Datta: https://github.com/Sumalyo

.. _Susanna Aro: https://www.linkedin.com/in/susanna-aro

.. _Svea Marie Meyer: https://github.com/SveaMeyer13

.. _T. Wang: https://github.com/twang5

.. _Tal Linzen: http://tallinzen.net/

.. _Teon Brooks: https://teonbrooks.com

.. _Theodore Papadopoulo: https://github.com/papadop

.. _Thomas Hartmann: https://github.com/thht

.. _Thomas Radman: https://github.com/tradman

.. _Timothy Gates: https://au.linkedin.com/in/tim-gates-0528a4199

.. _Tod Flak: https://github.com/todflak

.. _Tommy Clausner: https://github.com/TommyClausner

.. _Toomas Erik Anijärv: http://www.toomaserikanijarv.com/

.. _Tristan Stenner: https://github.com/tstenner/

.. _Tziona NessAiver: https://github.com/TzionaN

.. _Valerii Chirkov: https://github.com/vagechirkov

.. _Victor Ferat: https://github.com/vferat

.. _Victoria Peterson: https://github.com/vpeterson

.. _Xiaokai Xia: https://github.com/dddd1007

.. _Yaroslav Halchenko: http://haxbylab.dartmouth.edu/ppl/yarik.html

.. _Yousra Bekhti: https://www.linkedin.com/pub/yousra-bekhti/56/886/421

.. _Yu-Han Luo: https://github.com/yh-luo

.. _Zhi Zhang: https://github.com/tczhangzhi/

.. _Dinara Issagaliyeva: https://github.com/dissagaliyeva

.. _Jennifer Behnke: https://github.com/JKBehnke

.. _Hakimeh Pourakbari: https://github.com/Hpakbari

.. _Pavel Navratil: https://github.com/navrpa13

.. _Tom Ma: https://github.com/myd7349
