<table class="table table-hover table-striped table-sm table-responsive small">
    <tr>
        <th>Channels</th>
        <td>{{ channels }}</td>
    </tr>
    <tr>
        <th>Source space</th>
        <td>{{ source_space_descr }}</td>
    </tr>
    <tr>
        <th>Source orientation</th>
        <td>{{ source_orientation }}</td>
    </tr>
</table>
