.. This (-*- rst -*-) format file contains commonly used link targets
   and name substitutions.  It may be included in many files,
   therefore it should only contain link targets and name
   substitutions.  Try grepping for "^\.\. _" to find plausible
   candidates for this list.

.. NOTE: reST targets are
   __not_case_sensitive__, so only one target definition is needed for
   nipy, NIPY, Nipy, etc...


.. mne

.. _`MNE-Python`: http://mne.tools/mne-python-intro
.. _`MNE-Python GitHub`: https://github.com/mne-tools/mne-python
.. _`mne-scripts`: https://github.com/mne-tools/mne-scripts/
.. _`MNE-C manual`: https://mne.tools/mne-c-manual/MNE-manual-2.7.3.pdf
.. _`GitHub issues page`: https://github.com/mne-tools/mne-python/issues/
.. _`MNE Forum`: https://mne.discourse.group
.. _`MNE-BIDS`: https://mne.tools/mne-bids
.. _`MNE-BIDS-Pipeline`: https://mne.tools/mne-bids-pipeline
.. _`MNE-HCP`: http://mne.tools/mne-hcp
.. _`MNE-Realtime`: https://mne.tools/mne-realtime
.. _`MNE-MATLAB git repository`: https://github.com/mne-tools/mne-matlab
.. _`MNE-Docker`: https://github.com/mne-tools/mne-docker
.. _OpenMEEG: http://openmeeg.github.io
.. _EOSS2: https://chanzuckerberg.com/eoss/proposals/improving-usability-of-core-neuroscience-analysis-tools-with-mne-python
.. _EOSS4: https://chanzuckerberg.com/eoss/proposals/building-pediatric-and-clinical-data-pipelines-for-mne-python/
.. _`code of conduct`: https://github.com/mne-tools/.github/blob/main/CODE_OF_CONDUCT.md


.. TUTORIAL LINKS

.. _errors: https://en.wikipedia.org/w/index.php?title=Type_I_and_type_II_errors#Table_of_error_types
.. _fwer: https://en.wikipedia.org/wiki/Family-wise_error_rate
.. _fdr: https://en.wikipedia.org/wiki/False_discovery_rate
.. _ft_cluster: http://www.fieldtriptoolbox.org/faq/how_not_to_interpret_results_from_a_cluster-based_permutation_test
.. _ft_cluster_effect_size: https://mailman.science.ru.nl/pipermail/fieldtrip/2017-September/024644.html
.. _ft_exch: https://mailman.science.ru.nl/pipermail/fieldtrip/2008-October/001794.html



.. TEMPLATE FOR LINKS TO OTHER PROJECTS
.. _PROJECTNAME: http://neuroimaging.scipy.org
.. _`PROJECTNAME github`: http://github.com/nipy
.. _`PROJECTNAME mailing list`: http://projects.scipy.org/mailman/listinfo/nipy-devel

.. numpy

.. _numpy: http://www.numpy.org
.. _`numpy github`: http://github.com/numpy/numpy
.. _`numpy mailing list`: http://mail.scipy.org/mailman/listinfo/numpy-discussion


.. scipy

.. _scipy: http://www.scipy.org
.. _`scipy github`: http://github.com/scipy/scipy
.. _`scipy mailing list`: http://mail.scipy.org/mailman/listinfo/scipy-dev

.. freesurfer

.. _freesurfer: https://surfer.nmr.mgh.harvard.edu/
.. _`freesufer github`: https://github.com/freesurfer/freesurfer
.. _`freesurfer mailing list`: https://surfer.nmr.mgh.harvard.edu/fswiki/FreeSurferSupport


.. nipy

.. _nipy: http://nipy.org/nipy
.. _`nipy github`: http://github.com/nipy/nipy
.. _`nipy mailing list`: http://mail.scipy.org/mailman/listinfo/nipy-devel


.. ipython

.. _ipython: https://ipython.org
.. _`ipython github`: https://github.com/ipython/ipython
.. _`ipython mailing list`: https://mail.python.org/mailman/listinfo/ipython-dev


.. dipy

.. _dipy: http://nipy.org/dipy
.. _`dipy github`: http://github.com/Garyfallidis/dipy
.. _`dipy mailing list`: http://mail.scipy.org/mailman/listinfo/nipy-devel


.. nibabel

.. _nibabel: http://nipy.org/nibabel
.. _`nibabel github`: http://github.com/nipy/nibabel
.. _`nibabel mailing list`: http://mail.scipy.org/mailman/listinfo/nipy-devel


.. h5py

.. _h5py: http://www.h5py.org
.. _`h5py github`: http://github.com/h5py/h5py
.. _`h5py mailing list`: https://groups.google.com/forum/#!forum/h5py


.. Pytest

.. _pytest: https://docs.pytest.org/


.. Flake8

.. _Flake8: http://flake8.pycqa.org/


.. pymatreader

.. _pymatreader: https://gitlab.com/obob/pymatreader


.. h5io

.. _h5io: https://github.com/h5io/h5io


.. CuPy

.. _CuPy: https://cupy.chainer.org/

.. Dask

.. _Dask: https://dask.org/

.. git stuff

.. _git: https://git-scm.com/
.. _github: https://github.com
.. _GitHub Help: https://help.github.com
.. _GitHub learning lab: https://lab.github.com/
.. _pro git book: https://git-scm.com/book/
.. _git bash: https://gitforwindows.org/
.. _git for Windows: https://gitforwindows.org/
.. _git clone: http://schacon.github.com/git/git-clone.html
.. _git checkout: https://schacon.github.io/git/git-checkout.html
.. _git commit: https://schacon.github.io/git/git-commit.html
.. _git push: https://schacon.github.io/git/git-push.html
.. _git pull: https://schacon.github.io/git/git-pull.html
.. _git add: https://schacon.github.io/git/git-add.html
.. _git status: https://schacon.github.io/git/git-status.html
.. _git diff: https://schacon.github.io/git/git-diff.html
.. _git log: https://schacon.github.io/git/git-log.html
.. _git branch: https://schacon.github.io/git/git-branch.html
.. _git remote: https://schacon.github.io/git/git-remote.html
.. _git rebase: https://schacon.github.io/git/git-rebase.html
.. _git config: https://schacon.github.io/git/git-config.html

.. other stuff

.. _python: http://www.python.org
.. _Brain Imaging Data Structure: https://bids.neuroimaging.io/

.. python packages

.. _pep8: https://pypi.org/project/pep8/
.. _pyflakes: https://pypi.org/project/pyflakes
.. _coverage: https://pypi.python.org/pypi/coverage
.. _mayavi: https://docs.enthought.com/mayavi/mayavi/
.. _nitime: http://nipy.org/nitime/
.. _joblib: https://pypi.python.org/pypi/joblib
.. _scikit-learn: https://scikit-learn.org/stable/
.. _pyDICOM: https://pypi.python.org/pypi/pydicom/
.. _matplotlib: https://matplotlib.org/
.. _sphinx: https://www.sphinx-doc.org/
.. _pandas: https://pandas.pydata.org/
.. _PIL: https://pypi.python.org/pypi/PIL
.. _tqdm: https://tqdm.github.io/
.. _pooch: https://www.fatiando.org/pooch/latest/

.. python editors

.. _Spyder: https://www.spyder-ide.org/
.. _`integrated development environment`: https://en.wikipedia.org/wiki/Integrated_development_environment
.. _`spyder`: https://www.spyder-ide.org/
.. _`visual studio code`: https://code.visualstudio.com/
.. _`pycharm`: https://www.jetbrains.com/pycharm/

.. _anaconda: https://www.anaconda.com/products/individual
.. _miniconda: https://conda.io/en/latest/miniconda.html
.. _miniforge: https://github.com/conda-forge/miniforge
.. _installation instructions for Anaconda: http://docs.continuum.io/anaconda/install
.. _installation instructions for Miniconda: https://conda.io/projects/conda/en/latest/user-guide/install/index.html
.. _Anaconda troubleshooting guide: http://conda.pydata.org/docs/troubleshooting.html

.. installation links

.. _requirements file:  https://raw.githubusercontent.com/mne-tools/mne-python/main/requirements.txt
.. _NVIDIA CUDA GPU processing: https://developer.nvidia.com/cuda-zone
.. _NVIDIA proprietary drivers: https://www.geforce.com/drivers

.. _Sphinx documentation: https://www.sphinx-doc.org/en/master/usage/restructuredtext/basics.html
.. _sphinx-gallery: https://sphinx-gallery.github.io
.. _NumPy docstring style guidelines: https://numpydoc.readthedocs.io/en/latest/format.html#docstring-standard
.. _Stack Overflow: https://stackoverflow.com/

.. vim: ft=rst
