.. _changes_0_1:

Version 0.1 (2011-09-23)
------------------------

- Initial release (initial commit: 2010-12-26).

Authors
~~~~~~~

The committer list for this release is the following (preceded by number
of commits):

* 315  Alexandre Gramfort
*   1  Ellen Lau
*  41  Emily Ruzich
*   2  Martin Luessi
*   9  Matti Hämäläinen
