#container {
    position: relative
    padding-bottom: 5rem
}

#content {
    margin-top: 90px
    scroll-behavior: smooth
    position: relative  // for scrollspy
}

#toc {
    margin-top: 90px
    padding-bottom: 5rem
}

footer {
    margin-top: 5rem;
}
